/* ======================================================================
     This is a part of "gpop" - Gaussian output post processor
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
                                            created: Sept. 19, 2005
       - added forceV0 option to hindered rotor      Aug.  11, 2009
       - added coord in resultTST                    Oct.   8, 2009
       - added GauChebGrd in tempList                Dec.   8, 2009
 ====================================================================== */

#include "gpoplibt.h"

/*------------------------------------------------------------------------
     internal rotor description for modInfoTST
------------------------------------------------------------------------*/

/*----- an internal rotor ----------------------------------------------*/

intRotor::intRotor() { idxVib = -1; }             //---- (constructor) ---

void intRotor::clear() {                          //------------ clear ---
  idxVib = -1; isAuto = false; nSym = 0; redB = 0.; isFreeRotor = false;
  forceV0 = false; V0 = 0.;
  rotAtms.clear(); cntAtms.clear(); descript.erase();
}

                                                  //---------- setFrom ---
int intRotor::setFrom(int iv, int ns, vector<string> &mi, double V0in,
 string &ds, jobInfos &ji, molecGeom &mg, freqTable &ft) {
  double frq;
  vector<pair<int, double> > simv;
  clear();
  idxVib = iv - 1; nSym = ns;
  if (mbm.setList(mi, mg) == ERRE) { return ERRE; }
  mg.geom.setPrcAxesMomInert();
  mbm.setProp(mg);
  redB = mbm[0].asmRedB;
  rotAtms = mbm[0].lstAtms; cntAtms = mbm[1].lstAtms;
  descript = ds;
  mbm.setIntRotDevn(mg);
  if (idxVib < 0) {
    simv = ft.getMostAlike(mbm.intRotDevn);
    idxVib = simv[0].first; csim = simv[0].second; isAuto = true;
  } else {
    csim = fabs(mbm.intRotDevn * ft[idxVib].nrm); isAuto = false;
  }
  frq = ft[idxVib].freq * ji.scaleFactZPE;
  if (fabs(V0in) <= 1e-10) {
    // --- V0 = 0. means a free rotor
    V0 = 0.; isFreeRotor = true; forceV0 = true;
  } else if (V0in >= 1e-10) {
    // --- V0 > 0. means a hindered rotor with input V0
    V0 = V0in; isFreeRotor = false; forceV0 = true;
  } else {
    V0 = (frq / nSym) * (frq / nSym) / redB;
    isFreeRotor = false; forceV0 = false;
  }
  return NORM;
}

int intRotor::restoreFrom(istream &is) {          //------ restoreFrom ---
  int i, n;
  string slin;
  vector<string> tkns, aitk;
  clear();
  if (getLineGpop(is, slin) == ERRE) { return ERRE; }
  tkns = splitTokens(slin, " ");
  if (tkns.size() < 8) { return ERRE; }
  idxVib = atoi(tkns[0].c_str()) - 1;
  nSym = atoi(tkns[1].c_str());
  redB = atof(tkns[2].c_str());
  isFreeRotor = getBlnvStr(tkns[3]);
  forceV0 = getBlnvStr(tkns[4]);
  V0 = atof(tkns[5].c_str());
  aitk = splitTokens(tkns[6], "-"); n = aitk.size();
  for (i = 0; i < n; i++) { rotAtms.push_back(atoi(aitk[i].c_str()) - 1); }
  aitk = splitTokens(tkns[7], "-"); n = aitk.size();
  for (i = 0; i < n; i++) { cntAtms.push_back(atoi(aitk[i].c_str()) - 1); }
  if (tkns.size() > 8) { descript = tkns[8]; }
  return NORM;
}

ostream &intRotor::print(ostream &os) {           //------------ print ---
  os << "    " << (idxVib + 1) << " " << nSym;
  os.precision(6); os << " " << redB;
  os << " " << strGpoBlnv(isFreeRotor);
  os << " " << strGpoBlnv(forceV0);
  os.precision(4); os << " " << V0;
  os << " " << strListIdx(rotAtms) << " " << strListIdx(cntAtms);
  os << " " << descript << endl;
  return os;
}

ostream &operator<<(ostream &os, intRotor &ir)
  { return ir.print(os); }

/*----- list of internal rotors ----------------------------------------*/

                                                  //------ restoreFrom ---
int listIntRotors::restoreFrom(istream &is, int nir) {
  int iir;
  intRotor ir;
  clear();
  for (iir = 0; iir < nir; iir++) {
    if (ir.restoreFrom(is) == ERRE) { return ERRE; }
    push_back(ir);
  }
  return NORM;
}

ostream &listIntRotors::print(ostream &os) {      //------------ print ---
  int iir, nir = size();
  for (iir = 0; iir < nir; iir++) { os << (*this)[iir]; }
  return os;
}

int listIntRotors::isIntRotor(int iv) {           //------- isIntRotor ---
  int iir, nir = size();
  for (iir = 0; iir < nir; iir++)
    { if ((*this)[iir].idxVib == iv) { return true; } }
  return false;
}

int listIntRotors::getIdxIntRotor(int iv) {       //--- getIdxIntRotor ---
  int iir, nir = size();
  for (iir = 0; iir < nir; iir++)
    { if ((*this)[iir].idxVib == iv) { return iir; } }
  return -1;
}

                                                  //---- findDupVibIdx ---
int listIntRotors::findDupVibIdx(vector<int> &irdup,
 map<int, set<int> > &duplst) {
  int ir, nr = size(), iv, retv = false;
  map<int, int> usedIdcs;
  set<int> idrw;
  usedIdcs.clear(); irdup.clear(); irdup.resize(nr, false);
  duplst.clear();
  for (ir = 0; ir < nr; ir++) {
    iv = (*this)[ir].idxVib;
    if (usedIdcs.find(iv) != usedIdcs.end()) {
      retv = true; irdup[ir] = true; irdup[usedIdcs[iv]] = true;
      if (duplst.find(iv) == duplst.end()) {
        idrw.clear(); idrw.insert(ir); idrw.insert(usedIdcs[iv]);
        duplst.insert(make_pair(iv, idrw));
      } else { duplst[iv].insert(ir); }
    } else { usedIdcs.insert(make_pair(iv, ir)); }
  }
  return retv;
}

                                               //--- diagnIntRotAssign ---
void listIntRotors::diagnIntRotAssign(freqTable &ft) {
  int ir, nr = size(), ivr, foundDup, ird[2], imx;
  string ws;
  vector<int> idHRdup;
  map<int, set<int> > dupList;
  map<int, set<int> >::iterator p;
  set<int>::iterator q;
  vector<pair<int, double> > modr;
  devnInfo mixdev[2];
  foundDup = findDupVibIdx(idHRdup, dupList);
  if (foundDup) {
    cout << "******************************\n";
    cout << "* CAUTION: duplicated idxVib *\n";
    cout << "******************************\n";
  }
  for (ir = 0; ir < nr; ir++) {
    cout << "IntRot[" << (ir + 1) << "] " << (*this)[ir].rotAtms[0]
     << "-@ " << (*this)[ir].cntAtms[0] << "-@ " << (*this)[ir].descript;
    if ((*this)[ir].isAuto) { cout << " (AUTO) "; }
    else { cout << " (stat) "; }
    cout << "mode= " << ((*this)[ir].idxVib + 1) << " csim= "
     << (*this)[ir].csim << endl;
    if ((*this)[ir].csim < .85) { cout << "  Warning: small csim.\n"; }
    if (idHRdup[ir]) {
      cout << "  Caution: idxVib duplication. Other candidates are:\n";
      modr = ft.getMostAlike((*this)[ir].mbm.intRotDevn);
      for (ivr = 0; ivr < modr.size(); ivr++) {
        cout << "    mode= " << (modr[ivr].first + 1) << " : csim= "
          << modr[ivr].second << endl;
      }
    }
  }
  for (p = dupList.begin(); p != dupList.end(); p++) {
    if ((p->second).size() == 2) {
      cout << "\n***** Decoupling for duplication of mode= "
       << (p->first + 1) << " *****\n";
      q = (p->second).begin(); ird[0] = *q; q++; ird[1] = *q;
      mixdev[0] = (*this)[ird[0]].mbm.intRotDevn
       + (*this)[ird[1]].mbm.intRotDevn;
      mixdev[1] = (*this)[ird[0]].mbm.intRotDevn
       - (*this)[ird[1]].mbm.intRotDevn;
      for (imx = 0; imx < 2; imx++) {
        mixdev[imx].normalize();
        modr = ft.getMostAlike(mixdev[imx]);
        if (imx == 0) { ws = "  sum of "; }
        else { ws = "  diff of "; }
        cout << ws << "IntRot[" << (ird[0] + 1) << "] and IntRot["
         << (ird[1] + 1) << "] resembles:\n";
        for (ivr = 0; ivr < modr.size(); ivr++) {
          cout << "    mode= " << (modr[ivr].first + 1) << " : csim= "
            << modr[ivr].second << endl;
        }
      }
    }
  }
}

ostream &operator<<(ostream &os, listIntRotors &lir) //--- operator << ---
  { return lir.print(os); }

/*------------------------------------------------------------------------
     TST related modification informations
------------------------------------------------------------------------*/

modInfoTST::modInfoTST() { isTS = false; }        //---- (constructor) ---

void modInfoTST::clear()                          //------------ clear ---
 { isTS = false; hndRots.clear(); eStates.clear(); isomStates.clear(); }

int modInfoTST::restoreFrom(istream &is) {        //------ restoreFrom ---
  int numIR, numES, numIS;
  string key;
  vector<string> keytoks;

  clear();
  if (!getlGpoMatch(is, "tstInfo{")) { return ERRE; }
  if (!getlGpoBlnv(is, isTS)) { return ERRE; }
  if (!getlGpoDblv(is, energyTST)) { return ERRE; }

   //--- entries "intRotors", "eStates", and "isomStates" may be omitted
  if (!getlGpoKeyToks(is, key, keytoks)) { return ERRE; }
  if (key == "intRotors{") {
    if (keytoks.size() < 1) { return ERRE; }
    numIR = atoi(keytoks[0].c_str());
    if (hndRots.restoreFrom(is, numIR) == ERRE) { return ERRE; }
    if (!getlGpoMatch(is, "}intRotors")) { return ERRE; }
    if (!getlGpoKeyToks(is, key, keytoks)) { return ERRE; }
  }
  if (key == "eStates{") {
    if (keytoks.size() < 1) { return ERRE; }
    numES = atoi(keytoks[0].c_str());
    if (eStates.restoreFrom(is, numES) == ERRE) { return ERRE; }
    if (!getlGpoMatch(is, "}eStates")) { return ERRE; }
    if (!getlGpoKeyToks(is, key, keytoks)) { return ERRE; }
  }
  if (key == "isomStates{") {
    if (keytoks.size() < 1) { return ERRE; }
    numIS = atoi(keytoks[0].c_str());
    if (isomStates.restoreFrom(is, numIS) == ERRE) { return ERRE; }
    if (!getlGpoMatch(is, "}isomStates")) { return ERRE; }
    if (!getlGpoKeyToks(is, key, keytoks)) { return ERRE; }
  }
  if (key != "}tstInfo") { return ERRE; }
  return NORM;
}

ostream &modInfoTST::print(ostream &os) {         //------------ print ---
  int nrt, nes, nis, prec;
  prec = os.precision(); os.setf(ios::fixed);
  nrt = hndRots.size(); nes = eStates.size(); nis = isomStates.size();
  os << "tstInfo{\n";
  os << "  isTS " << strGpoBlnv(isTS) << endl;
  os.precision(8); os << "  energyTST " << energyTST << endl;

   //--- skip output "intRotors", "eState", or "isomStates" when null
  if (nrt > 0) {
    os << "  intRotors{ " << nrt << endl << hndRots << "  }intRotors\n";
  }
  if (nes > 0)
    { os << "  eStates{ " << nes << endl << eStates << "  }eStates\n"; }
  if (nis > 0) {
    os << "  isomStates{ " << nis << endl << isomStates
       << "  }isomStates\n";
  }
  os << "}tstInfo\n";
  os.unsetf(ios::fixed); os.precision(prec);
  return os;
}

ostream &operator<<(ostream &os, modInfoTST &mit)
  { return mit.print(os); }

/*------------------------------------------------------------------------
     List of temperatures
------------------------------------------------------------------------*/

tempList::tempList() { setDefault(); }            //---- (constructor) ---

void tempList::setDefault()                       //------- setDefault ---
  { setByRange(300., 1500., 100.); }

int tempList::readFrom(string flnam) {            //--------- readFrom ---
  string slin, key;
  vector<string> tkns;
  ifstream inf;
  inf.open(flnam.c_str());
  if (!inf) { return errms(FAILopen, flnam); }
  if (getLineGpop(inf, slin) == NORM) {
    if (setBy(slin) == ERRE) { inf.close(); return ERRE; }
    while (getLineGpop(inf, slin) == NORM) {
      if (addFrom(slin) == ERRE) { inf.close(); return ERRE; }
    }
  } else {
    return errms(EMPfile, flnam);
  }
  inf.close();
  return NORM;
}

int tempList::addFrom(string slin) {   // --- add from an input string ---
  string key;
  vector<string> tkns;
  tkns = splitTokens(slin, " ,\t");
  if (tkns.empty()) { return errms(INVparN); }
  key = tkns[0];
  if (key == "tempRange") { return addRangeVS(tkns); }
  else if (key == "tempRecipRange") { return addRecipRangeVS(tkns); }
  else if (key == "tempGauChebGrd") { return addGauChebGrdVS(tkns); }
  else if (key == "tempList") { return addListVS(tkns); }
  return errms(INVkey, key);
}

int tempList::setBy(string slin)         // --- set by an input string ---
 { clear(); return addFrom(slin); }

                          // --- sort and eliminate duplicated element ---
void tempList::sortElimDup() {
  int id, jd, nelm = size();
  double d, dlast;
  vector<double> vd = *this;
  if (nelm < 2) { return; }
  for (id = 0; id < (nelm - 1); id++) {
    for (jd = (id + 1); jd < nelm; jd++) {
      if (vd[jd] < vd[id]) { d = vd[id]; vd[id] = vd[jd]; vd[jd] = d; }
    }
  }
  clear(); push_back(vd[0]); dlast = vd[0];
  for (id = 1; id < nelm; id++) {
    if ((vd[id] - dlast) > 1.e-10) { push_back(vd[id]); dlast = vd[id]; }
  }
}

                                    // --- add a range of temperatures ---
int tempList::addRange(double st, double et, double dt) {
  double t;
  if (dt < 1.e-6) { return errms(INVstep); }
  if ((st < 1.e-6) || (et < st)) { return errms(INVrange); }
  if (((et - st) / dt) > 1.e4) { return errms(INVrngMANY); }
  for (t = st; t < (et + dt * .1); t += dt) { push_back(t); }
  sortElimDup();
  return NORM;
}

                                       // --- add a range from strings ---
int tempList::addRangeVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempRange"
  double st, et, dt;
  if (vs.size() < 4) { return errms(INVparN); }
  st = atof(vs[1].c_str()); et = atof(vs[2].c_str());
  dt = atof(vs[3].c_str());
  return addRange(st, et, dt);
}

                                 // --- set by a range of temperatures ---
int tempList::setByRange(double st, double et, double dt)
 { clear(); return addRange(st, et, dt); }

                                    // --- set by a range from strings ---
int tempList::setByRangeVS(vector<string> &vs)
 { clear(); return addRangeVS(vs); }

                            // --- add a range from reciprocal numbers ---
int tempList::addRecipRange(double numr, double stt, double lst,
 double stp) {
  double rt;
  if (stp < 1.e-12) { return errms(INVstep); }
  if ((stt < 1.e-12) || (lst < stt)) { return errms(INVrange); }
  if (((lst - stt) / stp) > 1.e4) { return errms(INVrngMANY); }
  for (rt = stt; rt < (lst + stp * .1); rt += stp)
   { push_back(numr / rt); }
  sortElimDup();
  return NORM;
}

                 // --- add a range from reciprocal numbers in strings ---
int tempList::addRecipRangeVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempRecipRange"
  double numr, stt, lst, stp;
  if (vs.size() < 5) { return errms(INVparN); }
  numr = atof(vs[1].c_str()); stt = atof(vs[2].c_str());
  lst = atof(vs[3].c_str()); stp = atof(vs[4].c_str());
  return addRecipRange(numr, stt, lst, stp);
}

                         // --- set by a range from reciprocal numbers ---
int tempList::setByRecipRange(double numr, double stt, double lst,
 double stp)
 { clear(); return addRecipRange(numr, stt, lst, stp); }

              // --- set by a range from reciprocal numbers in strings ---
int tempList::setByRecipRangeVS(vector<string> &vs)
 { clear(); return addRecipRangeVS(vs); }

                                // --- add Gauss-Chebyshev grid points ---
int tempList::addGauChebGrd(double Tmin, double Tmax, int nT) {
  int i;
  double Ttil, Tinv, Tmini = 1. / Tmin, Tmaxi = 1. / Tmax;
  for (i = 0; i < nT; i++) {
    Ttil = cos((2. * i + 1.) / 2. / nT * M_PI);
    Tinv = (Ttil * (Tmaxi - Tmini) + (Tmaxi + Tmini)) / 2.;
    push_back(1. / Tinv);
  }
  sortElimDup();
  return NORM;
}

                     // --- add Gauss-Chebyshev grid points in strings ---
int tempList::addGauChebGrdVS(vector<string> &vs) {
  int nT;
  double Tmin, Tmax;
  if (vs.size() < 4) { return errms(INVparN); }
  Tmin = atof(vs[1].c_str()); Tmax = atof(vs[2].c_str());
  nT = atoi(vs[3].c_str());
  return addGauChebGrd(Tmin, Tmax, nT);
}

                             // --- set by Gauss-Chebyshev grid points ---
int tempList::setByGauChebGrd(double Tmin, double Tmax, int nT)
 { clear(); return addGauChebGrd(Tmin, Tmax, nT); }

                  // --- set by Gauss-Chebyshev grid points in strings ---
int tempList::setByGauChebGrdVS(vector<string> &vs)
 { clear(); return addGauChebGrdVS(vs); }

                                     // --- add a list of temperatures ---
int tempList::addList(vector<double> &vd) {
  int id, nd = vd.size();
  for (id = 0; id < nd; id++) {
    if (vd[id] < 1.e-6) { return errms(INVtemp); }
    push_back(vd[id]);
  }
  sortElimDup();
  return NORM;
}

                          // --- add a list of temperatures in strings ---
int tempList::addListVS(vector<string> &vs) {
  // - assumes vs[0] is key, "tempList"
  int is, ns = vs.size();
  vector<double> vd;
  vd.clear();
  for (is = 1; is < ns; is++) { vd.push_back(atof(vs[is].c_str())); }
  if (!vd.empty()) { return addList(vd); }
  return NORM;
}

                                  // --- set by a list of temperatures ---
int tempList::setByList(vector<double> &vd)
 { clear(); return addList(vd); }

                       // --- set by a list of temperatures in strings ---
int tempList::setByListVS(vector<string> &vs)
 { clear(); return addListVS(vs); }

ostream &tempList::print(ostream &os) {           //------------ print ---
  int itl, ntl = size();
  os << "  tempList";
  for (itl = 0; itl < ntl; itl++) { os << " " << (*this)[itl]; }
  os << endl;
  return os;
}

double tempList::maxTemp() {                      //---------- maxTemp ---
  int itl, ntl = size();
  double maxT = 0.;
  for (itl = 0; itl < ntl; itl++) {
    if ((*this)[itl] > maxT) { maxT = (*this)[itl]; }
  }
  return maxT;
}

                                                  // --- error message ---
int tempList::errms(errcode erc, string sv) {
  string msg = "tempList: ";
  switch(erc) {
    case FAILopen: msg += "Failed to open [" + sv + "]."; break;
    case EMPfile: msg += "Empty file [" + sv + "]."; break;
    case INVkey: msg += "Invalid key[" + sv + "]."; break;
    case INVstep: msg += "Invalid (too small) step input."; break;
    case INVrange: msg += "Invalid range input."; break;
    case INVrngMANY: msg += "Invalid range (too many) input."; break;
    case INVparN: msg += "Too small number of input parameters."; break;
    case INVtemp: msg += "Invalid temperature (< 0) input."; break;
  }
  cout << msg << endl;
  return ERRE;
}

ostream &operator<<(ostream &os, tempList &tl)    //------ operator << ---
  { return tl.print(os); }

/*------------------------------------------------------------------------
     molecular internal mode
------------------------------------------------------------------------*/

molIntMode::molIntMode() { clear(); }             //---- (constructor) ---

void molIntMode::clear() {                        //------------ clear ---
  type = Vibrator; nSym = 1; forceV0 = false; freq = 0.; redB = 0.;
  V0 = 0.;
}

ostream &molIntMode::print(ostream &os) {         //------------ print ---
  os << "    ";
  switch (type) {
    case Vibrator: os << "Vibrator "; break;
    case FreeRotor: os << "FreeRotor "; break;
    case HindRotor: os << "HindRotor "; break;
    default: os << "? "; break;
  }
  os << nSym;
  os.precision(4); os << " " << freq;
  os.precision(6); os << " " << redB << endl;
  return os;
}

double molIntMode::modeQ(double T) {              //------------ modeQ ---
  double qmode;
  switch(type) {
    case Vibrator: qmode = qHOvib(T, freq); break;
    case FreeRotor:
     qmode = q1DRRrot(T, redB, nSym);
     if (forceV0) { qmode *= qLostZPE(T, freq); }
     break;
    case HindRotor: qmode = qPGHindRot(T, freq, redB, nSym, forceV0, V0);
     break;
    default: qmode = 1.; break;
  }
  return qmode;
}

double molIntMode::modeU(double T) {              //------------ modeU ---
  double umode;
  switch(type) {
    case Vibrator: umode = uHOvib(T, freq); break;
    case FreeRotor:
     umode = u1DRRrot(T, redB, nSym);
     if (forceV0) { umode += uLostZPE(T, freq); }
     break;
    case HindRotor: umode = uPGHindRot(T, freq, redB, nSym, forceV0, V0);
     break;
    default: umode = 0.; break;
  }
  return umode;
}

double molIntMode::modeS(double T) {              //------------ modeS ---
  double smode;
  switch(type) {
    case Vibrator: smode = sHOvib(T, freq); break;
    case FreeRotor: smode = s1DRRrot(T, redB, nSym); break;
    case HindRotor: smode = sPGHindRot(T, freq, redB, nSym, forceV0, V0);
     break;
    default: smode = 0.; break;
  }
  return smode;
}

double molIntMode::modeCv(double T) {             //----------- modeCv ---
  double cvmode;
  switch(type) {
    case Vibrator: cvmode = cvHOvib(T, freq); break;
    case FreeRotor: cvmode = cv1DRRrot(T, redB, nSym); break;
    case HindRotor: cvmode = cvPGHindRot(T, freq, redB, nSym, forceV0, V0);
     break;
    default: cvmode = 0.; break;
  }
  return cvmode;
}

                                                  //------ operator << ---
ostream &operator<<(ostream &os, molIntMode &minm)
  { return minm.print(os); }

     /* ---------- global thermodynamic functions --------------------- */

                                                      //--- qLostZPE
double qLostZPE(double T, double freq) {
  double x = freq * KELWAVN / T;
  return exp(x / 2.);
}
                                                      //--- uLostZPE
double uLostZPE(double T, double freq) {
  double x = freq * KELWAVN / T;
  return -x / 2.;
}
                                   /* --- hind. rot. Pitzer-Gwinn ----- */

                                                      //--- qPGHindRot
double qPGHindRot(double T, double freq, double B, int sgm,
 int forceV0, double V0) {
  if (!forceV0) { V0 = freq * freq / (B * sgm * sgm); }
  return qClHindRot(T, B, sgm, V0) * qHOvib(T, freq) / qClHOvib(T, freq);
}

                                                      //--- uPGHindRot
double uPGHindRot(double T, double freq, double B, int sgm,
 int forceV0, double V0) {
  if (!forceV0) { V0 = freq * freq / (B * sgm * sgm); }
  return uClHindRot(T, B, sgm, V0) + uHOvib(T, freq) - uClHOvib(T, freq);
}

                                                      //--- sPGHindRot
double sPGHindRot(double T, double freq, double B, int sgm,
 int forceV0, double V0) {
  if (!forceV0) { V0 = freq * freq / (B * sgm * sgm); }
  return sClHindRot(T, B, sgm, V0) + sHOvib(T, freq) - sClHOvib(T, freq);
}

                                                      //-- cvPGHindRot
double cvPGHindRot(double T, double freq, double B, int sgm,
 int forceV0, double V0) {
  if (!forceV0) { V0 = freq * freq / (B * sgm * sgm); }
  return cvClHindRot(T, B, sgm, V0) + cvHOvib(T, freq)
   - cvClHOvib(T, freq);
}

                                   /* --- vibrational HO/quantum ------ */

double qHOvib(double T, double freq) {                 //------ qHOvib
  double x = freq * KELWAVN / T;
  return 1. / (1. - exp(-x));
}

double uHOvib(double T, double freq) {                 //------ uHOvib
  double x = freq * KELWAVN / T;
  return x / (exp(x) - 1.);
}

double sHOvib(double T, double freq) {                 //------ sHOvib
  double x = freq * KELWAVN / T;
  return x / (exp(x) - 1.) - log(1. - exp(-x));
}

double cvHOvib(double T, double freq) {                //----- cvHOvib
  double x = freq * KELWAVN / T, ex;
  ex = exp(x);
  return ex * pow(x / (ex - 1.), 2);
}

                                   /* --- vibrational HO/classical ---- */

double qClHOvib(double T, double freq) {               //---- qClHOvib
  double x = freq * KELWAVN / T;
  return 1. / x;
}

double uClHOvib(double T, double freq) { return 1.; }  //---- uClHOvib

double sClHOvib(double T, double freq) {               //---- sClHOvib
  double x = freq * KELWAVN / T;
  return 1. - log(x);
}

double cvClHOvib(double T, double freq) { return 1.; } //--- cvClHOvib

                                   /* --- hindered rotor classical ---- */

                                                      //--- qClHindRot
double qClHindRot(double T, double B, int sgm, double V0) {
  double y = V0 * KELWAVN / 2. / T;
  return q1DRRrot(T, B, sgm) * exp(-y) * besselI(0, y);
}

                                                      //--- uClHindRot
double uClHindRot(double T, double B, int sgm, double V0) {
  double y = V0 * KELWAVN / 2. / T;
  return 0.5 + y * (1. - besselI(1, y) / besselI(0, y));
}

                                                      //--- sClHindRot
double sClHindRot(double T, double B, int sgm, double V0) {
  double y = V0 * KELWAVN / 2. / T;
  return 0.5 + y * (1. - besselI(1, y) / besselI(0, y))
   + log(qClHindRot(T, B, sgm, V0));
}

                                                      //-- cvClHindRot
double cvClHindRot(double T, double B, int sgm, double V0) {
  double y = V0 * KELWAVN / 2. / T, bI0, bI1, bI2;
  bI0 = besselI(0, y); bI1 = besselI(1, y); bI2 = besselI(2, y);
  return 0.5 + y * y * (bI2 * bI0 + bI1 * bI0 / y - bI1 * bI1)
   / (bI0 * bI0);
}

                                   /* --- 1-D free rotor -------------- */

double q1DRRrot(double T, double B, int sgm)           //---- q1DRRrot
  { return sqrt(M_PI * T / (B * KELWAVN)) / sgm; }

double u1DRRrot(double T, double B, int sgm)           //---- u1DRRrot
  { return 0.5; }

double s1DRRrot(double T, double B, int sgm)           //---- s1DRRrot
  { return 0.5 + log(q1DRRrot(T, B, sgm)); }

double cv1DRRrot(double T, double B, int sgm)          //--- cv1DRRrot
  { return 0.5; }

double besselI(int n, double x) {                 //---------- besselI ---
  int i = 1;
  double retv = 0., q = x * x / 4., tm;
  if (x < 0.) { return retv; }
  retv = (tm = pow(x / 2., n) / factorial(n));
  if (q <= 1e-20) { return retv; }
  do {
    tm *= q / i / (i + n); retv += tm; i++;
  } while ((tm / retv) > 1e-10);
  return retv;
}

double factorial(int n) {                         //-------- factorial ---
  int i;
  double retv = 1.;
  if (n < 1) { return retv; }
  for (i = n; i > 0; i--) { retv *= i; }
  return retv;
}

/*------------------------------------------------------------------------
     thermofunc calculation results
------------------------------------------------------------------------*/

/*----- Thermofunc calculation results at a temperature ----------------*/

void res1Therm::printCSV(ostream &os) {           //--------- printCSV ---
  int prec;
  prec = os.precision(); os.setf(ios::fixed);
  os.precision(2); os << T << ",";
  os.precision(4); os << S << "," << Cp << "," << H << endl;
  os.unsetf(ios::fixed); os.precision(prec);
}

                                                  //---- printCSVsubtH ---
void res1Therm::printCSVsubtH(ostream &os, double Hr) {
  int prec;
  prec = os.precision(); os.setf(ios::fixed);
  os.precision(2); os << T << ",";
  os.precision(4); os << S << "," << Cp << "," << (H - Hr) << endl;
  os.unsetf(ios::fixed); os.precision(prec);
}

/*----- Thermofunc calculation for a molecule --------------------------*/

void resMolTherm::printHeadCSV(ostream &os)       //----- printHeadCSV ---
  { os << "T[K],S[JK-1mol-1],Cp[JK-1mol-1],H[kJmol-1]\n"; }

void resMolTherm::printHeadCSVs(ostream &os)      //---- printHeadCSVs ---
  { os << "T[K],S[JK-1mol-1],Cp[JK-1mol-1],H-H298[kJmol-1]\n"; }

                                                  //---- printCSVsubtH ---
void resMolTherm::printCSVsubtH(ostream &os) {
  int irm, nrm = size();
  for (irm = 0; irm < nrm; irm++)
    { (*this)[irm].printCSVsubtH(os, ref.H); }
}

/*------------------------------------------------------------------------
     molecular properties for TST calculation
------------------------------------------------------------------------*/

molPropTST::molPropTST() { clear(); }             //---- (constructor) ---

void molPropTST::clear() {                        //------------ clear ---
  int i;
  molecType = Invalid; molnam.erase(); rotSymNbr = -1; sigma_nIsom = -1.;
  molW = 0.;
  for (i = 0; i < 3; i++) { rotC[i] = 0.; }
  energyTST = 0.; imagFreqTS = 0.;
  intMode.clear(); eStates.clear(); isomStates.clear();
}

int molPropTST::readFromTST(string basfln) {      //------ readFromTST ---
  int gel, nes, nis, nrc, irc, nvb, ivb, nhr, ihr;
  double freq;
  string tstfln = basfln + ".tst", comm, slin;
  vector<string> tkns;
  ifstream inf;
  molIntMode minm;

  clear();
  inf.open(tstfln.c_str());
  if (!inf)
    { cout << "failed to open [" << tstfln << "].\n"; return ERRE; }
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  comm = slin;
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  molW = atof(slin.c_str());
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  gel = atoi(slin.c_str());
  if (gel < 0) {
    nes = -gel;
    if (eStates.restoreFromTST(inf, nes) == ERRE) { return ERRE; }
  } else if (gel > 0) { eStates.setOneState(gel); }
  else { return ERRE; }
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  sigma_nIsom = atof(slin.c_str());
  if (sigma_nIsom < 0.) {
    if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
    tkns = splitTokens(slin, " ");
    if (tkns.size() < 2) { return ERRE; }
    rotSymNbr = atoi(tkns[0].c_str()); nis = atoi(tkns[1].c_str());
    if (rotSymNbr <= 0) { return ERRE; }
    if (isomStates.restoreFromTST(inf, nis) == ERRE) { return ERRE; }
  } else { isomStates.setOneState(1); rotSymNbr = -1; }
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  nrc = atoi(slin.c_str());
  for (irc = 0; irc < nrc; irc++) {
    if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
    rotC[irc] = atof(slin.c_str()) / GHZWAVN;
  }
  switch (nrc) {
    case 0: molecType = Atom; break;
    case 1:
    case 2: molecType = Linear; break;
    default: molecType = NonLinear; break;
  }
  if (getLine(inf, slin) == ERRE) { return ERRE; }
  if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
  nvb = atoi(slin.c_str());
  for (ivb = 0; ivb < nvb; ivb++) {
    if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
    freq = atof(slin.c_str());
    minm.clear();
    if (freq < 0.)
      { minm.type = molIntMode::FreeRotor; minm.redB = -freq; }
    else { minm.type = molIntMode::Vibrator; minm.freq = freq; }
    intMode.push_back(minm);
  }
  if (getLine(inf, slin) == ERRE) { return NORM; }
  if (getLineTrim(inf, slin) == ERRE) { return NORM; }
  nhr = atoi(slin.c_str());
  for (ihr = 0; ihr < nhr; ihr++) {
    if (getLineTrim(inf, slin) == ERRE) { return ERRE; }
    tkns = splitTokens(slin, " ");
    minm.clear();
    minm.type = molIntMode::HindRotor; minm.redB = atof(tkns[0].c_str());
    minm.freq = atof(tkns[1].c_str()); minm.nSym = atoi(tkns[2].c_str());
    intMode.push_back(minm);
  }
  if (getLine(inf, slin) == ERRE) { return NORM; }
  if (getLineTrim(inf, slin) == ERRE) { return NORM; }
  energyTST = atof(slin.c_str());
  if (getLine(inf, slin) == ERRE) { return NORM; }
  if (getLineTrim(inf, slin) == ERRE) { return NORM; }
  imagFreqTS = fabs(atof(slin.c_str()));
  molnam = basfln;
  inf.close();
  return NORM;
}

                                                  //---------- setFrom ---
void molPropTST::setFrom(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, modInfoTST &mit) {
  int gel, nis, irc, nrc, ifr0 = 0, ifr, nfr, ihr;
  molIntMode minm;

  clear();
  if (mg.nAtms == 1) { molecType = Atom; nrc = 0; }
  else if (mg.isLinear) { molecType = Linear; nrc = 1; }
  else { molecType = NonLinear; nrc = 3; }
  molnam = basfln; gel = mg.gElec; molW = mg.molW;
  nis = (int)mg.numIsomers; rotSymNbr = (int)mg.rotSymNbr;
  switch (molecType) {
    case NonLinear:
      for (irc = 0; irc < nrc; irc++)
        { rotC[irc] = mg.rotcGHz[irc] / GHZWAVN; }
      break;
    case Linear:
      rotC[0] = mg.rotcGHz[1] / GHZWAVN;
      break;
    default: break;
  }
  energyTST = mit.energyTST;
  if (mit.isTS) { imagFreqTS = fabs(ft[0].freq); ifr0 = 1; }
  nfr = ft.size();
  for (ifr = ifr0; ifr < nfr; ifr++) {
    minm.clear();
    if ((ihr = mit.hndRots.getIdxIntRotor(ifr)) >= 0) {
      if (mit.hndRots[ihr].isFreeRotor) {
        minm.type = molIntMode::FreeRotor;
      } else {
        minm.type = molIntMode::HindRotor;
      }
      minm.nSym = mit.hndRots[ihr].nSym;
      minm.forceV0 = mit.hndRots[ihr].forceV0;
      minm.freq = ft[ifr].freq * ji.scaleFactZPE;
      minm.redB = mit.hndRots[ihr].redB;
      minm.V0 = mit.hndRots[ihr].V0;
    } else {
      minm.type = molIntMode::Vibrator;
      minm.freq = ft[ifr].freq * ji.scaleFactVib;
    }
    intMode.push_back(minm);
  }
  eStates = mit.eStates;
  if (eStates.size() < 1) { eStates.setOneState(gel); }
  isomStates = mit.isomStates;
  if (isomStates.size() < 1) { isomStates.setOneState(nis); }
}

int molPropTST::numVib() {                                //--- numVib ---
  int imim, nmim = intMode.size(), nVib = 0;
  for (imim = 0; imim < nmim; imim++) {
    if (intMode[imim].type == molIntMode::Vibrator) { ++nVib; }
  }
  return nVib;
}

int molPropTST::numIntRot() {                          //--- numIntRot ---
  int imim, nmim = intMode.size(), nIRot = 0;
  for (imim = 0; imim < nmim; imim++) {
    if ( (intMode[imim].type == molIntMode::FreeRotor) ||
         (intMode[imim].type == molIntMode::HindRotor) )
      { ++nIRot; }
  }
  return nIRot;
}

ostream &molPropTST::print(ostream &os) {         //------------ print ---
  int prec, irc, nrc = 0, imim, nmim, nes, nis;

  prec = os.precision(); os.setf(ios::fixed);
  os << "molPropTST{\n  molecType ";
  switch (molecType) {
    case Invalid: os << "Invalid\n"; break;
    case Atom: os << "Atom\n"; break;
    case Linear: os << "Linear\n"; nrc = 1; break;
    case NonLinear: os << "NonLinear\n"; nrc = 3; break;
    default: os << "?\n"; break;
  }
  os << "  molnam " << molnam << endl;
  os << "  rotSymNbr " << rotSymNbr << endl;
  os.precision(10); os << "  sigma_nIsom " << sigma_nIsom << endl;
  os.precision(6); os << "  molW " << molW << endl;
  os.precision(8); os << "  rotC";
  for (irc = 0; irc < nrc; irc++) { os << " " << rotC[irc]; }
  os << "\n  energyTST " << energyTST << endl;
  os.precision(4); os << "  imagFreqTS " << imagFreqTS << endl;
  nmim = intMode.size(); nes = eStates.size(); nis = isomStates.size();
  os << "  intMode{ " << nmim << endl;
  for (imim = 0; imim < nmim; imim++) { os << intMode[imim]; }
  os << "  }intMode\n";
  os << "  eStates{ " << nes << endl << eStates << "  }eStates\n";
  os << "  isomStates{ " << nis << endl << isomStates
     << "  }isomStates\n";
  os << "}molPropTST\n";
  os.unsetf(ios::fixed); os.precision(prec);
  return os;
}

     /* ---------- thermodynamic functions ---------------------------- */

                                                  //--------- getTherm ---
void molPropTST::getTherm(double T, double &S, double &Cp, double &H) {
  S = totalS1bar(T) * RGASC;                   // J K-1 mol-1
  Cp = (totalCv(T) + 1.) * RGASC;              // J K-1 mol-1
  H = (totalU(T) + 1.) * RGASC * T / 1000.;    // kJ mol-1
}

                                                  //------ getMolTherm ---
resMolTherm molPropTST::getMolTherm(tempList &tl) {
  int itl, ntl = tl.size();
  resMolTherm res;
  res.clear();
  res.ref = get1Therm(STDTMPR);
  for (itl = 0; itl < ntl; itl++) { res.push_back(get1Therm(tl[itl])); }
  return res;
}

res1Therm molPropTST::get1Therm(double T) {       //-------- get1Therm ---
  res1Therm res;
  res.T = T;
  res.S = totalS1bar(T) * RGASC;                  // J K-1 mol-1
  res.Cp = (totalCv(T) + 1.) * RGASC;             // J K-1 mol-1
  res.H = (totalU(T) + 1.) * RGASC * T / 1000.;   // kJ mol-1
  return res;
}

double molPropTST::totalQ(double T) {                  //------ totalQ
  double qV, qFR, qHR;
  return transQ(T) * elecQ(T) * isomQ(T) * rotQ(T)
         * imodQ(T, qV, qFR, qHR);
}

double molPropTST::totalU(double T) {                  //------ totalU
  double uV, uFR, uHR;
  return transU(T) + elecU(T) + isomU(T) + rotU(T)
         + imodU(T, uV, uFR, uHR);
}

double molPropTST::totalS1bar(double T) {              //-- totalS1bar
  double sV, sFR, sHR;
  return transS1bar(T) + elecS(T) + isomS(T) + rotS(T)
         + imodS(T, sV, sFR, sHR);
}

double molPropTST::totalCv(double T) {                 //----- totalCv
  double cvV, cvFR, cvHR;
  return transCv(T) + elecCv(T) + isomCv(T) + rotCv(T)
         + imodCv(T, cvV, cvFR, cvHR);
}

                                   /* ---------- translational -------- */

double molPropTST::transQ(double T)                    //------ transQ
  { return CTRANS * pow(molW * T, 1.5); }

double molPropTST::transU(double T) { return 1.5; }    //------ transU

double molPropTST::transS1bar(double T)                //-- transS1bar
  { return 1.5 * log(molW) + 2.5 * log(T) + CSTR1B; }

double molPropTST::transCv(double T) { return 1.5; }   //----- transCv

                                   /* ---------- electronic ----------- */

double molPropTST::elecQ(double T) { return eStates.statesQ(T); }
double molPropTST::elecU(double T) { return eStates.statesU(T); }
double molPropTST::elecS(double T) { return eStates.statesS(T); }
double molPropTST::elecCv(double T) { return eStates.statesCv(T); }

                                   /* ---------- isomer-contribution -- */

double molPropTST::isomQ(double T) { return isomStates.statesQ(T); }
double molPropTST::isomU(double T) { return isomStates.statesU(T); }
double molPropTST::isomS(double T) { return isomStates.statesS(T); }
double molPropTST::isomCv(double T) { return isomStates.statesCv(T); }

                                   /* ---------- extern-rotational ---- */

double molPropTST::rotQ(double T) {                    //-------- rotQ
  int irt;
  double qrot, sigma, kBTB = 1.;
  if (rotSymNbr < 0) { sigma = sigma_nIsom; } else { sigma = rotSymNbr; }
  switch (molecType) {
    case Linear:
      qrot = T / (rotC[0] * KELWAVN) / sigma; break;
    case NonLinear:
      for (irt = 0; irt < 3; irt++) { kBTB *= T / (rotC[irt] * KELWAVN); }
      qrot = sqrt(M_PI * kBTB) / sigma; break;
    default:
      qrot = 1.; break;
  }
  return qrot;
}

double molPropTST::rotU(double T) {                    //-------- rotU
  double urot;
  switch (molecType) {
    case Linear:    urot = 1.;  break;
    case NonLinear: urot = 1.5; break;
    default:        urot = 0.;  break;
  }
  return urot;
}

double molPropTST::rotS(double T) {                    //-------- rotS
  double srot;
  switch (molecType) {
    case Linear:    srot = 1. + log(rotQ(T));  break;
    case NonLinear: srot = 1.5 + log(rotQ(T)); break;
    default:        srot = 0.;  break;
  }
  return srot;
}

double molPropTST::rotCv(double T) {                   //------- rotCv
  double cvrot;
  switch (molecType) {
    case Linear:    cvrot = 1.;  break;
    case NonLinear: cvrot = 1.5; break;
    default:        cvrot = 0.;  break;
  }
  return cvrot;
}

                                                  //------------ imodQ ---
double molPropTST::imodQ(double T, double &qV, double &qFR, double &qHR) {
  int imim, nmim = intMode.size();
  qV = 1.; qFR = 1.; qHR = 1.;
  for (imim = 0; imim < nmim; imim++) {
    switch(intMode[imim].type) {
      case molIntMode::Vibrator: qV *= intMode[imim].modeQ(T); break;
      case molIntMode::FreeRotor: qFR *= intMode[imim].modeQ(T); break;
      case molIntMode::HindRotor: qHR *= intMode[imim].modeQ(T); break;
      default: break;
    }
  }
  return qV * qFR * qHR;
}

                                                  //------------ imodU ---
double molPropTST::imodU(double T, double &uV, double &uFR, double &uHR) {
  int imim, nmim = intMode.size();
  uV = 0.; uFR = 0.; uHR = 0.;
  for (imim = 0; imim < nmim; imim++) {
    switch(intMode[imim].type) {
      case molIntMode::Vibrator: uV += intMode[imim].modeU(T); break;
      case molIntMode::FreeRotor: uFR += intMode[imim].modeU(T); break;
      case molIntMode::HindRotor: uHR += intMode[imim].modeU(T); break;
      default: break;
    }
  }
  return uV + uFR + uHR;
}

                                                  //------------ imodS ---
double molPropTST::imodS(double T, double &sV, double &sFR, double &sHR) {
  int imim, nmim = intMode.size();
  sV = 0.; sFR = 0.; sHR = 0.;
  for (imim = 0; imim < nmim; imim++) {
    switch(intMode[imim].type) {
      case molIntMode::Vibrator: sV += intMode[imim].modeS(T); break;
      case molIntMode::FreeRotor: sFR += intMode[imim].modeS(T); break;
      case molIntMode::HindRotor: sHR += intMode[imim].modeS(T); break;
      default: break;
    }
  }
  return sV + sFR + sHR;
}

                                                  //----------- imodCv ---
double molPropTST::imodCv(double T, double &cvV, double &cvFR,
 double &cvHR) {
  int imim, nmim = intMode.size();
  cvV = 0.; cvFR = 0.; cvHR = 0.;
  for (imim = 0; imim < nmim; imim++) {
    switch(intMode[imim].type) {
      case molIntMode::Vibrator: cvV += intMode[imim].modeCv(T); break;
      case molIntMode::FreeRotor: cvFR += intMode[imim].modeCv(T); break;
      case molIntMode::HindRotor: cvHR += intMode[imim].modeCv(T); break;
      default: break;
    }
  }
  return cvV + cvFR + cvHR;
}

ostream &operator<<(ostream &os, molPropTST &mpt) //------ operator << ---
  { return mpt.print(os); }

/*----- molecular information set for reaction specification -----------*/

molInfoSet::molInfoSet() { clear(); }             //---- (constructor) ---

void molInfoSet::clear() {                        //------------ clear ---
  status = Invalid; isTS = false; isUM = false;
  energy = 0.; imagFreqTS = 0.;
  basnams.clear(); molp.clear();
}

int molInfoSet::readFrom(istream &is) {           //--------- readFrom ---
  int endBlock = false, ntkns, itkn;
  string slin, key;
  vector<string> tkns;

  do {
    if (getLineGpop(is, slin) == ERRE) { return ERRE; }
    tkns = splitTokens(slin, " "); key = tkns[0]; ntkns = tkns.size();
    if (key == "}") {
      endBlock = true;
    } else if (key == "file") {
      if (ntkns < 2)
        { cout << "No argument for [file].\n"; return ERRE; }
      if (isTS) {
        if ((ntkns > 2) || (basnams.size() > 0))
          { cout << "Only one file allowed for TS.\n"; return ERRE; }
      }
      if (isUM) {
        if ((ntkns > 2) || (basnams.size() > 0)) {
          cout << "Only one file allowed for UNIMOL reactant.\n"; return ERRE;
        }
      }
      for (itkn = 1; itkn < ntkns; itkn++)
        { basnams.push_back(tkns[itkn]); }
      status |= File;
    } else if (key == "energy") {
      if (ntkns < 2)
        { cout << "No argument for [energy].\n"; return ERRE; }
      else if ((ntkns > 2) || (status & Energy) != 0)
        { cout << "Too many energies.\n"; return ERRE; }
      energy = atof(tkns[1].c_str());
      status |= Energy;
    } else if (key == "imagFreqTS") {
      if (!isTS) {
        cout << "Keyword [imagFreqTS] is allowed only for TS.\n";
        return ERRE;
      } else if ((status & ImagFreq) != 0) {
        cout << "Too many imaginary frequencies.\n"; return ERRE;
      }
      imagFreqTS = fabs(atof(tkns[1].c_str()));
      status |= ImagFreq;
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return ERRE;
    }
  } while (!endBlock);
  return NORM;
}

int molInfoSet::readProp(string ftyp) {           //--------- readProp ---
  int ipfl, npfl = basnams.size();
  double sumEnergy = 0.;
  molPropTST mptst;
  jobInfos jbinf;
  molecGeom molc;
  freqTable frtab;
  modInfoTST miTST;

  if ((status & File) != 0) {
    for (ipfl = 0; ipfl < npfl; ipfl++) {
      mptst.clear();
      if (ftyp == "tst") {
        if (mptst.readFromTST(basnams[ipfl]) == ERRE) { return ERRE; }
        molp.push_back(mptst);
      } else if (ftyp == "gpo") {
        if (readGpo(basnams[ipfl], jbinf, molc, frtab, miTST) == ERRE)
          { return ERRE; }
        mptst.setFrom(basnams[ipfl], jbinf, molc, frtab, miTST);
        molp.push_back(mptst);
      } else { return ERRE; }
      sumEnergy += mptst.energyTST;
    }
    if ((status & Energy) == 0)
      { energy = sumEnergy; status |= Energy; }
    if (isTS && ((status & ImagFreq) == 0))
      { imagFreqTS = mptst.imagFreqTS; status |= ImagFreq; }
  }
  return NORM;
}

ostream &molInfoSet::print(ostream &os) {         //------------ print ---
  int prec, ibn, nbn, imp, nmp;

  prec = os.precision(); os.setf(ios::fixed);
  os << "molInfoSet{\n  status " << status << endl;
  os << "  isTS " << strGpoBlnv(isTS) << endl;
  os.precision(10); os << "  energy " << energy << endl;
  os.precision(4); os << "  imagFreqTS " << imagFreqTS << endl;
  nbn = basnams.size(); nmp = molp.size();
  os << "  basnams";
  for (ibn = 0; ibn < nbn; ibn++) { os << " " << basnams[ibn]; }
  os << "\n  molp{ " << nmp << "\n!------------------------------\n";
  os.unsetf(ios::fixed); os.precision(prec);
  for (imp = 0; imp < nmp; imp++) { os << molp[imp]; }
  os << "!------------------------------\n  }molp\n";
  os << "}molInfoSet\n";
  return os;
}

double molInfoSet::molsetQ(double T) {            //---------- molsetQ ---
  int imp, nmp = molp.size();
  double qset = 1.;
  for (imp = 0; imp < nmp; imp++) { qset *= molp[imp].totalQ(T); }
  return qset;
}

double molInfoSet::molsetU(double T) {            //---------- molsetU ---
  int imp, nmp = molp.size();
  double uset = 0.;
  for (imp = 0; imp < nmp; imp++) { uset += molp[imp].totalU(T); }
  return uset;
}

double molInfoSet::molsetS1bar(double T) {        //------ molsetS1bar ---
  int imp, nmp = molp.size();
  double sset = 0.;
  for (imp = 0; imp < nmp; imp++) { sset += molp[imp].totalS1bar(T); }
  return sset;
}

double molInfoSet::molsetCv(double T) {           //--------- molsetCv ---
  int imp, nmp = molp.size();
  double cvset = 0.;
  for (imp = 0; imp < nmp; imp++) { cvset += molp[imp].totalCv(T); }
  return cvset;
}

double molInfoSet::molsetH(double T) {            //---------- molsetH ---
  int imp, nmp = molp.size();
  double hset = 0.;
  for (imp = 0; imp < nmp; imp++) { hset += (molp[imp].totalU(T) + 1); }
  return hset;
}

double molInfoSet::molsetG1bar(double T) {        //------ molsetG1bar ---
  int imp, nmp = molp.size();
  double gset = 0.;
  for (imp = 0; imp < nmp; imp++)
    { gset += (molp[imp].totalU(T) + 1 - molp[imp].totalS1bar(T)); }
  return gset;
}

                                                  //------ operator << ---
ostream &operator<<(ostream &os, molInfoSet &mis)
  { return mis.print(os); }

/*----- TST calculation results at a temperature -----------------------*/

resultTST::resultTST() { clear(); }               //---- (constructor) ---

void resultTST::clear() {                         //------------ clear ---
  T = 0.; kTST = 0.; Keq = 0.; corTunSh = 0.; corTunEck = 0.; coord = 0.;
}

ostream &resultTST::print(ostream &os) {          //------------ print ---
  int prec;
  os << "    " << T; prec = os.precision(); os.precision(8);
  os << " " << kTST << " " << Keq
     << " " << corTunSh << " " << corTunEck << endl;
  os.precision(prec);
  return os;
}

ostream &operator<<(ostream &os, resultTST &rT)   //------ operator << ---
  { return rT.print(os); }

/*----- tunneling correction -------------------------------------------*/

tunnelCorr::tunnelCorr() { clear(); }             //---- (constructor) ---

void tunnelCorr::clear() {                        //------------ clear ---
  flgShavitt = false; flgEckart = false; vMax = 0.; vDiff = 0.; fImg = 0.;
  vB = 0.; beta = 0.; tpcoshD = 0.; eneMax = 0.; eneGrain = 0.;
  maxNumError = 0.;
}

                                                  //------ initShavitt ---
void tunnelCorr::initShavitt(double v0, double fi) {
  vMax = v0; fImg = fi;
  if (vMax < 0.) { return; }
  flgShavitt = true;
}

                                                  //------- initEckart ---
void tunnelCorr::initEckart(double v0, double vD, double fi,
 double maxT) {
  vMax = v0; vDiff = vD; fImg = fi;
  if ((vMax < 0.) || (vMax < vDiff)) { return; }
  vB = pow(sqrt(vMax) + sqrt(vMax - vDiff), 2);
  beta = sqrt(vB / (2 * vMax * (vMax - vDiff))) * fImg / (2 * M_PI);
  tpcoshD = cosh(sqrt(2 * vB - pow(M_PI * beta, 2)) / beta);
  eneMax = maxT / KELWAVN * 30. + vMax; eneGrain = 10.;
  while ((eneMax / eneGrain) < 1.e+3) { eneGrain /= 10.; }
  eneMin = eneGrain / 10.;
  if (vDiff > eneMin) { eneMin += vDiff; }
  flgShavitt = true; flgEckart = true;
}

ostream &tunnelCorr::print(ostream &os) {         //------------ print ---
  int prec;
  prec = os.precision(); os.precision(8);
  os << "  tunnelCorr{\n    flgShavitt/flgEckart ";
  os << strGpoBlnv(flgShavitt) << "/" << strGpoBlnv(flgEckart) << endl;
  os << "    vMax/vDiff/fImg " << vMax << "/" << vDiff << "/" << fImg
     << endl;
  os << "    vB/beta/tpcoshD " << vB << "/" << beta << "/" << tpcoshD
     << endl;
  os << "    eneMin/eneMax/eneGrain " << eneMin << "/" << eneMax
     << "/" << eneGrain << endl;
  os << "    maxNumError " << maxNumError << endl;
  os << "  }tunnelCorr\n";
  os.precision(prec);
  return os;
}

double tunnelCorr::corrFacShavitt(double T) {     //--- corrFacShavitt ---
  double corrf = 1.;
  if (!flgShavitt) { return corrf; }
  corrf += pow(fImg * KELWAVN / T, 2) / 24. * (1. + T / (vMax * KELWAVN));
  return corrf;
}

double tunnelCorr::corrFacEckart(double T) {      //---- corrFacEckart ---
  double corrf = 1., cfClAn, cfClNm, cfQm, errCl;
  if (!flgEckart) { return corrf; }
  cfClAn = fClAnalytic(T); cfClNm = fClNumeric(T);
  cfQm = fQmNumeric(T);
  errCl = fabs((cfClNm - cfClAn) / cfClAn);
  if (errCl > maxNumError) { maxNumError = errCl; }
  corrf = cfQm / cfClAn;
  return corrf;
}

double tunnelCorr::fClAnalytic(double T) {        //------ fClAnalytic ---
  return T / KELWAVN * exp(-vMax * KELWAVN / T);
}

double tunnelCorr::fClNumeric(double T) {         //------- fClNumeric ---
  double fcl = 0., ene, f0, f1, f2, kT = T / KELWAVN;
  for (ene = vMax; ene < eneMax; ene += eneGrain) {
    f0 = exp(-ene / kT);
    f1 = exp(-(ene + eneGrain / 2.) / kT);
    f2 = exp(-(ene + eneGrain) / kT);
    fcl += (f0 + 4. * f1 + f2) * eneGrain / 6.;
  }
  return fcl;
}

double tunnelCorr::fQmNumeric(double T) {         //------- fQmNumeric ---
  double fqm = 0., ene, f0, f1, f2, kT = T / KELWAVN;
  for (ene = vMax; ene < eneMax; ene += eneGrain) {
    f0 = tpEck(ene) * exp(-ene / kT);
    f1 = tpEck(ene + eneGrain / 2.) * exp(-(ene + eneGrain / 2.) / kT);
    f2 = tpEck(ene + eneGrain) * exp(-(ene + eneGrain) / kT);
    fqm += (f0 + 4. * f1 + f2) * eneGrain / 6.;
  }
  for (ene = vMax; ene >= (eneMin + eneGrain); ene -= eneGrain) {
    f0 = tpEck(ene) * exp(-ene / kT);
    f1 = tpEck(ene - eneGrain / 2.) * exp(-(ene - eneGrain / 2.) / kT);
    f2 = tpEck(ene - eneGrain) * exp(-(ene - eneGrain) / kT);
    fqm += (f0 + 4. * f1 + f2) * eneGrain / 6.;
  }
  return fqm;
}

double tunnelCorr::tpEck(double E) {              //------------ tpEck ---
  double tpA, tpB, tpcoshApB, tpcoshAmB;
  tpA = sqrt(2. * E) / beta;
  tpB = sqrt(2. * (E - vDiff)) / beta;
  tpcoshApB = cosh(tpA + tpB); tpcoshAmB = cosh(tpA - tpB);
  return (tpcoshApB - tpcoshAmB) / (tpcoshApB + tpcoshD);
}

ostream &operator<<(ostream &os, tunnelCorr &tC)  //------ operator << ---
  { return tC.print(os); }

/*----- reaction information for TST calculation -----------------------*/

reactionInfo::reactionInfo() { clear(); }         //---- (constructor) ---

void reactionInfo::clear() {                      //------------ clear ---
  int i;
  calRat = false; calTun = false; calEck = false; calEql = false;
  fileType = "gpo"; baseName.erase(); enCnvFac = 1.;
  mReac.clear(); mTS.clear(); mProd.clear(); mTS.isTS = true;
  tnlC.clear(); calcRes.clear();
  dU0 = 0.; dH298 = 0.; dS298 = 0.;
  dUts0 = 0.; dHts298 = 0.; dSts298 = 0.;
  for (i = 0; i < 3; i++) { apA[i] = 0.; apER[i] = 0.; fmxErr[i] = 0.; }
}

int reactionInfo::readFrom(string basfln) {       //--------- readFrom ---
  int ntkns, tempRead = false;
  string rxnfln = basfln + ".rxn", slin, key, arg;
  vector<string> tkns;
  ifstream inf;

  clear();
  inf.open(rxnfln.c_str());
  if (!inf)
    { cout << "failed to open [" << rxnfln << "].\n"; return ERRE; }
  while (getLineGpop(inf, slin) == NORM) {
    tkns = splitTokens(slin, " "); key = tkns[0]; ntkns = tkns.size();
    if (key == "fileType") {
      if (ntkns < 2)
        { cout << "No argument for [fileType].\n"; return ERRE; }
      arg = tkns[1];
      if ((arg == "gpo") || (arg == "tst")) { fileType = arg; }
      else {
        cout << "Invalid argument [" << arg << "] for [fileType].\n";
        return ERRE;
      }
    } else if (key == "energyUnit") {
      if (ntkns < 2)
        { cout << "No argument for [energyUnit].\n"; return ERRE; }
      arg = tkns[1];
      if (arg == "hartree") { enCnvFac = 1.; }
      else if (arg == "kJ/mol") { enCnvFac = 1. / KJMLHRT; }
      else if (arg == "cm-1") { enCnvFac = 1. / WAVNHRT; }
      else {
        cout << "Invalid argument [" << arg << "] for [energyUnit].\n";
        return ERRE;
      }
    } else if (key.substr(0, 4) == "temp") {
      if (!tempRead) {
        if (temperat.setBy(slin) == ERRE) { return ERRE; }
        tempRead = true;
      } else {
        if (temperat.addFrom(slin) == ERRE) { return ERRE; }
      }
    } else if (key == "reactants{") {
      if (mReac.readFrom(inf) == ERRE) { return ERRE; }
    } else if (key == "transitionState{") {
      if (mTS.readFrom(inf) == ERRE) { return ERRE; }
    } else if (key == "products{") {
      if (mProd.readFrom(inf) == ERRE) { return ERRE; }
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return ERRE;
    }
  }
  inf.close();
  if ((mReac.status & Energy) != 0)
    { mReac.energy *= enCnvFac; }
  if ((mTS.status & Energy) != 0)
    { mTS.energy *= enCnvFac; }
  if ((mProd.status & Energy) != 0)
    { mProd.energy *= enCnvFac; }
  baseName = basfln;
  return NORM;
}

int reactionInfo::readProp() {                    //--------- readProp ---
  if (mReac.readProp(fileType) == ERRE) { return ERRE; }
  if (mTS.readProp(fileType) == ERRE) { return ERRE; }
  if (mProd.readProp(fileType) == ERRE) { return ERRE; }
  if ((mReac.status & File) == 0)                      // no reactant
    { cout << "No reactant file specified.\n"; return ERRE; }
  if ((mProd.status & File) != 0) { calEql = true; }
  if ((mTS.status & File) == 0) { return NORM; }       // no TS
  calRat = true;
  if (fabs(mTS.imagFreqTS) <= 1e-10) { return NORM; }
  calTun = true;
  if ((mProd.status & (Energy | File)) == 0) { return NORM; }
  calEck = true;
  return NORM;
}

ostream &reactionInfo::print(ostream &os) {       //------------ print ---
  int prec, icr, ncr;
  os << "reactionInfo{\n  calRat/Tun/Eck/Eql ";
  os << strGpoBlnv(calRat) << "/" << strGpoBlnv(calTun) << "/"
     << strGpoBlnv(calEck) << "/" << strGpoBlnv(calEql) << endl;
  os << "  fileType " << fileType << endl;
  os << "  baseName " << baseName << endl;
  prec = os.precision(); os.precision(10);
  os << "  enCnvFac " << enCnvFac << endl; os.precision(prec);
  os << temperat;
  os << "#===== Reactants ===========================================\n";
  os << mReac;
  os << "#===== Transition State ====================================\n";
  os << mTS;
  os << "#===== Products ============================================\n";
  os << mProd;
  os << "#===========================================================\n";
  os << tnlC;
  os << "  calcRes{\n";
  ncr = calcRes.size();
  for (icr = 0; icr < ncr; icr++) { os << calcRes[icr]; }
  os << "  }calcRes\n";
  return os;
}

void reactionInfo::calculate() {                  //-------- calculate ---
  int itl, ntl = temperat.size();
  double delEts, delHpr, imfTS, maxT, T, qReac, qTS, qProd;
  double hReac, hTS, hProd, sReac, sTS, sProd;
  resultTST rsT;

  delHpr = (mProd.energy - mReac.energy) * WAVNHRT;
  delEts = (mTS.energy - mReac.energy) * WAVNHRT;
  imfTS = mTS.imagFreqTS; maxT = temperat.maxTemp();
  if (calEck) { tnlC.initEckart(delEts, delHpr, imfTS, maxT); }
  else if (calTun) { tnlC.initShavitt(delEts, imfTS); }
  for (itl = 0; itl < ntl; itl++) {
    T = temperat[itl];
    rsT.clear(); rsT.T = T;
    if (calEql || calRat) { qReac = mReac.molsetQ(T); }
    if (calEql) {
      qProd = mProd.molsetQ(T);
      rsT.Keq = qProd / qReac * exp(-delHpr * KELWAVN / T);
    }
    if (calRat) {
      qTS = mTS.molsetQ(T);
      rsT.kTST = TSTK_H * T * qTS / qReac * exp(-delEts * KELWAVN / T);
      if (calTun) { rsT.corTunSh = tnlC.corrFacShavitt(T);
        if (calEck) { rsT.corTunEck = tnlC.corrFacEckart(T); }
      }
    }
    calcRes.push_back(rsT);
  }
  T = STDTMPR;
  if (calEql || calRat)
    { hReac = mReac.molsetH(T); sReac = mReac.molsetS1bar(T); }
  if (calEql) {
    dU0 = delHpr * JMLWAVN / 1000.;
    hProd = mProd.molsetH(T); sProd = mProd.molsetS1bar(T);
    dH298 = (hProd - hReac + delHpr * KELWAVN / T) * RGASC * T / 1000.;
    dS298 = (sProd - sReac) * RGASC;
  }
  if (calRat) {
    dUts0 = delEts * JMLWAVN / 1000.;
    hTS = mTS.molsetH(T); sTS = mTS.molsetS1bar(T);
    dHts298 = (hTS - hReac + delEts * KELWAVN / T) * RGASC * T / 1000.;
    dSts298 = (sTS - sReac) * RGASC;
  }
}

void reactionInfo::fitArrhenius() {               //----- fitArrhenius ---
  int i, nfit = 1, icr, ncr = calcRes.size();
  double x, y[3], Sx2 = 0., Sx = 0.;
  double Sxy[3] = {0., 0., 0.}, Sy[3] = {0., 0., 0.};
  double det, dlkmax[3] = {0., 0., 0.}, T, dlk;

  if (!calRat) { return; }
  if (calEck) { nfit = 3; }
  else if (calTun) { nfit = 2; }
  for (icr = 0; icr < ncr; icr++) {
    x = 1000. / calcRes[icr].T;
    y[0] = log(calcRes[icr].kTST);
    if (nfit > 1) { y[1] = y[0] + log(calcRes[icr].corTunSh); }
    if (nfit > 2) { y[2] = y[0] + log(calcRes[icr].corTunEck); }
    Sx2 += x * x; Sx += x;
    for (i = 0; i < nfit; i++) { Sxy[i] += x * y[i]; Sy[i] += y[i]; }
  }
  det = ncr * Sx2 - Sx * Sx;
  for (i = 0; i < nfit; i++) {
    apA[i] = exp((Sx2 * Sy[i] - Sx * Sxy[i]) / det);
    apER[i] = (Sx * Sy[i] - ncr * Sxy[i]) / det * 1000.;
  }
  for (icr = 0; icr < ncr; icr++) {
    T = calcRes[icr].T;
    y[0] = log(calcRes[icr].kTST);
    if (nfit > 1) { y[1] = y[0] + log(calcRes[icr].corTunSh); }
    if (nfit > 2) { y[2] = y[0] + log(calcRes[icr].corTunEck); }
    for (i = 0; i < nfit; i++) {
      dlk = fabs(y[i] - log(apA[i]) + apER[i] / T);
      if (dlk > dlkmax[i]) { dlkmax[i] = dlk; }
    }
  }
  for (i = 0; i < nfit; i++) { fmxErr[i] = exp(dlkmax[i]); }
}

int reactionInfo::outfileCSV() {                  //------- outfileCSV ---
  int prec, icr, ncr = calcRes.size();
  double T, ivT, kT, Ke, cSh, cEck;
  string csvfln;
  ofstream cof;

  csvfln = baseName + ".csv";
  cof.open(csvfln.c_str());
  if (!cof)
    { cout << "failed to open [" << csvfln << "].\n"; return ERRE; }
  prec = cof.precision(); cof.setf(ios::fixed);
  cof.precision(4);
  if (calEql) {
    cof << "dE0,kJmol-1," << dU0 << endl;
    cof << "dH298,kJmol-1," << dH298 << endl;
    cof << "dS298,JK-1mol-1," << dS298 << endl;
  }
  if (calRat) {
    cof << "dE*0,kJmol-1," << dUts0 << endl;
    cof << "dH*298,kJmol-1," << dHts298 << endl;
    cof << "dS*298,JK-1mol-1," << dSts298 << endl;
  }
  cof << endl;
  cof.unsetf(ios::fixed); cof.precision(7);
  cof << "T[K]";
  if (calEql) { cof << ",Keq"; }
  if (calRat) {
    if (calEck) { cof << ",k+Eck,corrEck,1000/T,ln(k+Eck)"; }
    if (calTun) { cof << ",k+Sh,corrSh,1000/T,ln(k+Sh)"; }
    cof << ",k,1000/T,ln(k)";
  }
  cof << endl;
  for (icr = 0; icr < ncr; icr++) {
    T = calcRes[icr].T; kT = calcRes[icr].kTST; Ke = calcRes[icr].Keq;
    cSh = calcRes[icr].corTunSh; cEck = calcRes[icr].corTunEck;
    ivT = 1000. / T;
    cof << T;
    if (calEql) { cof << "," << Ke; }
    if (calRat) {
      if (calEck) {
        cof << "," << (kT * cEck) << "," << cEck << "," << ivT
            << "," << log(kT * cEck);
      }
      if (calTun) {
        cof << "," << (kT * cSh) << "," << cSh << "," << ivT
            << "," << log(kT * cSh);
      }
      cof << "," << kT << "," << ivT << "," << log(kT);
    }
    cof << endl;
  }
  if (calRat) {
    cof << endl;
    if (calEck) { cof << ",k+Eck"; }
    if (calTun) { cof << ",k+Sh"; }
    cof << ",k\n";
    cof << "A";
    if (calEck) { cof << "," << apA[2]; }
    if (calTun) { cof << "," << apA[1]; }
    cof  << "," << apA[0] << endl;
    cof << "E/R";
    if (calEck) { cof << "," << apER[2]; }
    if (calTun) { cof << "," << apER[1]; }
    cof << "," << apER[0] << endl;
    cof << "ferr";
    if (calEck) { cof << "," << fmxErr[2]; }
    if (calTun) { cof << "," << fmxErr[1]; }
    cof << "," << fmxErr[0] << endl;
  }
  cof.precision(prec);
  cof.close();
  return NORM;
}

ostream &operator<<(ostream &os, reactionInfo &ri)
  { return ri.print(os); }

/*------------------------------------------------------------------------
     global library functions
------------------------------------------------------------------------*/

/*----- read gpop file into objects ------------------------------------*/

int readGpo(string &basfln, jobInfos &ji, molecGeom &mg, freqTable &ft,
 modInfoTST &mit) {
  string gpofln = basfln + ".gpo";
  ifstream inf;
  inf.open(gpofln.c_str());
  if (!inf)
    { cout << "failed to open [" << gpofln << "].\n"; return ERRE; }
  if (ji.restoreFrom(inf) == ERRE)
    { cout << "Error reading jobInfos.\n"; inf.close(); return ERRE; }
  if (mg.restoreFrom(inf) == ERRE)
    { cout << "Error reading molecGeom.\n"; inf.close(); return ERRE; }
  if (ft.restoreFrom(inf) == ERRE)
    { cout << "Error reading freqTable.\n"; inf.close(); return ERRE; }
  if (mit.restoreFrom(inf) == ERRE)
    { cout << "Error reading modInfoTST.\n"; inf.close(); return ERRE; }
  inf.close();
  return NORM;
}

/*----- output .gpo file -----------------------------------------------*/

int outFileGpo(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, modInfoTST &mit) {
  int prec;
  double eSCFzpe = atof(ji.SCFenergy.c_str()) + ft.zpe / WAVNHRT;
  string gpofln;
  ofstream gpof;
  gpofln = basfln + ".gpo";
  gpof.open(gpofln.c_str());
  if (!gpof)
    { cout << "failed to open [" << gpofln << "].\n"; return ERRE; }
  gpof << "# gpop link data file created from " << basfln << ".mod.\n";
  prec = gpof.precision(); gpof.setf(ios::fixed); gpof.precision(10);
  gpof << "# SCFenergy + zpe = " << eSCFzpe << endl;
  gpof.unsetf(ios::fixed); gpof.precision(prec);
  gpof << ji << mg << ft << mit;
  gpof.close();
  return NORM;
}

/*----- output .tst file -----------------------------------------------*/

//----- create .tst file from Gaussian output and .mod ---------

int outFileTst(string &basfln, jobInfos &ji, molecGeom &mg,
 freqTable &ft, modInfoTST &mit) {
  int irot, ifrq, numFreq, numHR, ihr, prec, nes, nis;
  string toutfln;
  ofstream tof;

  toutfln = basfln + ".tst";
  tof.open(toutfln.c_str());
  if (!tof)
    { cout << "failed to open [" << toutfln << "].\n"; return ERRE; }
  prec = tof.precision(); tof.setf(ios::fixed);
  tof << "[Name of the Molecule]\n";
  tof << " " << ji.comment << endl;
  tof << "[Molecular Mass(amu)]\n";
  tof.precision(6); tof << " " << mg.molW << endl;
  tof << "[Deg Electron State |neg|=NumState followed by (gi, ei(cm-1))'s]\n";
  if ((nes = mit.eStates.size()) > 0)
    { tof << " " << (-nes) << endl << mit.eStates.outStrTST(); }
  else { tof << " " << mg.gElec << endl; }
  tof << "[Rot Sym Num / Num of Isomrs]\n";
  tof << " -1\n " << (int)mg.rotSymNbr << " ";
  if ((nis = mit.isomStates.size()) > 0)
    { tof << nis << endl << mit.isomStates.outStrTST(); }
  else { tof << "1\n " << (int)mg.numIsomers << " 0.\n"; }
  tof << "[Num of RotConsts (1:linear 3:nonlinear), RotConsts(GHz)]\n";
  tof.precision(7);
  if (mg.nAtms == 1) {
    tof << " 0\n";
  } else {
    if (mg.isLinear) {
      tof << " 1\n";
      tof << " " << mg.rotcGHz[1] << endl;
    } else {
      tof << " 3\n";
      for (irot = 0; irot < 3; irot++)
        { tof << " " << mg.rotcGHz[irot] << endl; }
    }
  }
  tof << "[Num of Freqs, Freqs(cm-1) |neg|=1D-rotorB(cm-1)]\n";
  numFreq = mg.nVibm; numHR = mit.hndRots.size();
  if (mit.isTS) { numFreq--; }
  numFreq -= numHR;
  tof << " " << numFreq << endl;
  tof.precision(4);
  for (ifrq = 0; ifrq < mg.nVibm; ifrq++) {
    if ((!mit.isTS) || ifrq > 0) {
      if (!mit.hndRots.isIntRotor(ifrq))
        { tof << " " << ft[ifrq].freq * ji.scaleFactVib << endl; }
    }
  }
  tof << "[Num of H.R. / B(cm-1), Freq(cm-1), SigN]\n";
  tof << " " << numHR << endl;
  for (ihr = 0; ihr < numHR; ihr++) {
    tof.precision(6); tof << " " << mit.hndRots[ihr].redB;
    tof.precision(4); tof << " "
     << ft[mit.hndRots[ihr].idxVib].freq * ji.scaleFactZPE;
    tof << " " << mit.hndRots[ihr].nSym << endl;
  }
  tof << "[energy for TST (hartree)]\n";
  tof.precision(8); tof << " " << mit.energyTST << endl;
  if (mit.isTS) {
    tof << "[imaginary frequency (cm-1) .. only for TS molecule]\n";
    tof.precision(4); tof << " " << fabs(ft[0].freq) << endl;
  }
  tof.unsetf(ios::fixed); tof.precision(prec);
  tof.close();
  return NORM;
}

