/*========================================================================
     "vtst"  Variational TST calculation
             Copyright (c) 2002-2009 by A. Miyoshi, Univ. Tokyo
                                     created: Oct.   7, 2009
                                 last edited: Nov.  11, 2009
--------------------------------------------------------------------------
 usage: vtst basename
--------------------------------------------------------------------------
 - Read basename.rxv file. Also requires .gpo (or .tst) files specified
    in the .rxv file.
 - Keywords allowed in .rxv file are:
    fileType ftyp(s)       ! "gpo" (default) or "tst"
    energyUnit uStr(s)     ! "hartree" (default), "kJ/mol", or "cm-1"
    tempRange Tstart(double) Tend(double) Tstep(double)
    tempRecipRange numer(double) start(double) end(double) step(double)
    tempList T1(double) T2(double) ...
    reactants{
      file f1Re(s) [f2Re(s) ...]
      energy eRe(double)
    }
    products{
      file f1Pr(s) [f2Pr(s) ...]
      energy ePr(double)
    }
    varTS coord1(double) fTS1(s) eTS1(double)
    varTS coord2(double) fTS2(s) eTS2(double)
    :
========================================================================*/

#include "vtst.h"

/*===== main function ==================================================*/

int main(int argc, char *argv[]) {
  int its;
  string basfln, rxvfln, csvfln;
  ifstream inf;
  ofstream cof;
  reactionInfo *rx = new reactionInfo;
  listVTS *lvts = new listVTS;

  showRevision();
  if (argc != 2) { cout << "usage: vtst basename\n"; return 1; }
  basfln = argv[1];
  rxvfln = basfln + ".rxv";
  csvfln = basfln + ".csv";

  inf.open(rxvfln.c_str());
  if (!inf)
    { cout << "failed to open [" << rxvfln << "].\n"; return 1; }
  if (readRxv(inf, rx, lvts) == ERRE) { return 1; }
  inf.close();

  if (readPropRP(rx) == ERRE) { return 1; }

  for (its = 0; its < lvts->size(); its++) {
    rx->calcRes.clear();
    rx->mTS.clear(); rx->mTS.isTS = true;
    rx->mTS.basnams.push_back((*lvts)[its].flname);
    rx->mTS.status |= molInfoSet::File;
    if ((*lvts)[its].energySet) {
      rx->mTS.energy = (*lvts)[its].energy * rx->enCnvFac;
      rx->mTS.status |= molInfoSet::Energy;
    }
    if (rx->mTS.readProp(rx->fileType) == ERRE) { return 1; }
    rx->calculate();
    (*lvts)[its].calcRes = rx->calcRes; (*lvts)[its].dUts0 = rx->dUts0;
    (*lvts)[its].dHts298 = rx->dHts298; (*lvts)[its].dSts298 = rx->dSts298;
  }
  if (rx->calEql) {
    lvts->dU0 = rx->dU0; lvts->dH298 = rx->dH298; lvts->dS298 = rx->dS298;
  }

  if (findMinRate(rx, lvts) == ERRE) { return 1; }

  cof.open(csvfln.c_str());
  if (!cof)
    { cout << "failed to open [" << csvfln << "].\n"; return 1; }
  writeCsv(cof, lvts);
  cof.close();

  delete lvts, rx;
  return 0;
}

/*----- read-in .rxv file ----------------------------------------------*/

int readRxv(istream &is, reactionInfo *rx, listVTS *lvts) {
  int ntkns, tempRead = false;
  string slin, key, arg;
  vector<string> tkns;
  varTState vts;

  while (getLineGpop(is, slin) == NORM) {
    tkns = splitTokens(slin, " "); key = tkns[0]; ntkns = tkns.size();
    if (key == "fileType") {
      if (ntkns < 2)
        { cout << "No argument for [fileType].\n"; return ERRE; }
      arg = tkns[1];
      if ((arg == "gpo") || (arg == "tst")) { rx->fileType = arg; }
      else {
        cout << "Invalid argument [" << arg << "] for [fileType].\n";
        return ERRE;
      }
    } else if (key == "energyUnit") {
      if (ntkns < 2)
        { cout << "No argument for [energyUnit].\n"; return ERRE; }
      arg = tkns[1];
      if (arg == "hartree") { rx->enCnvFac = 1.; }
      else if (arg == "kJ/mol") { rx->enCnvFac = 1. / KJMLHRT; }
      else if (arg == "cm-1") { rx->enCnvFac = 1. / WAVNHRT; }
      else {
        cout << "Invalid argument [" << arg << "] for [energyUnit].\n";
        return ERRE;
      }
    } else if (key.substr(0, 4) == "temp") {
      if (!tempRead) {
        if (rx->temperat.setBy(slin) == ERRE) { return ERRE; }
        tempRead = true;
      } else {
        if (rx->temperat.addFrom(slin) == ERRE) { return ERRE; }
      }
    } else if (key == "reactants{") {
      if (rx->mReac.readFrom(is) == ERRE) { return ERRE; }
    } else if (key == "products{") {
      if (rx->mProd.readFrom(is) == ERRE) { return ERRE; }
    } else if (key == "varTS") {
      if (ntkns < 3) {
        cout << "Insufficient number of arguments for [varTS].\n";
        return ERRE;
      }
      vts.clear();
      vts.coord = atof(tkns[1].c_str());
      vts.flname = tkns[2];
      if (ntkns > 3) {
        vts.energySet = true;
        vts.energy = atof(tkns[3].c_str());
      }
      lvts->push_back(vts);
    } else {
      cout << "Invalid keyword [" << key << "].\n"; return ERRE;
    }
  }
  return NORM;
}

/*----- read-in property files for reactants and products --------------*/

int readPropRP(reactionInfo *rx) {
  if ((rx->mReac.status & molInfoSet::File) == 0)          // no reactant
    { cout << "No reactant file specified.\n"; return ERRE; }
  if ((rx->mProd.status & molInfoSet::File) != 0) { rx->calEql = true; }
  rx->calRat = true; rx->calTun = false; rx->calEck = false;
  if ((rx->mReac.status & molInfoSet::Energy) != 0)
    { rx->mReac.energy *= rx->enCnvFac; }
  if ((rx->mProd.status & molInfoSet::Energy) != 0)
    { rx->mProd.energy *= rx->enCnvFac; }
  if (rx->mReac.readProp(rx->fileType) == ERRE) { return ERRE; }
  if (rx->mProd.readProp(rx->fileType) == ERRE) { return ERRE; }
  return NORM;
}

/*----- find minimum rate along the coordinate -------------------------*/

int findMinRate(reactionInfo *rx, listVTS *lvts) {
  int its, itl;
  vector<double> x, y;
  spline3On s3;
  resultTST rtst;

  lvts->vtstRes.clear();
  lvts->calEql = rx->calEql;
  x.clear();
  for (its = 0; its < lvts->size(); its++)
    { x.push_back((*lvts)[its].coord); }
  for (itl = 0; itl < rx->temperat.size(); itl++) {
    y.clear(); s3.clear();
    for (its = 0; its < lvts->size(); its++)
      { y.push_back(log((*lvts)[its].calcRes[itl].kTST)); }
    if (s3.setFrom(x, y) == ERRE) { return ERRE; }
    rtst.kTST = exp(s3.findYmin(rtst.coord));
    rtst.T = rx->temperat[itl];
    if (lvts->calEql) { rtst.Keq = (*lvts)[0].calcRes[itl].Keq; }
    lvts->vtstRes.push_back(rtst);
  }
  return NORM;
}

/*----- write results to .csv file -------------------------------------*/

void writeCsv(ostream &os, listVTS *lvts) {
  int prec = os.precision(), icr, ncr = lvts->vtstRes.size();
  int its, nts = lvts->size();
  double T, kT, Ke, crd;

  os.setf(ios::fixed); os.precision(4);
  if (lvts->calEql) {
    os << "dE0,kJmol-1," << lvts->dU0 << endl;
    os << "dH298,kJmol-1," << lvts->dH298 << endl;
    os << "dS298,JK-1mol-1," << lvts->dS298 << endl;
  }
  os << "\ncoord,";
  for (its = 0; its < nts; its++) { os << "," << (*lvts)[its].coord; }
  os << "\ndE*0,kJmol-1";
  for (its = 0; its < nts; its++) { os << "," << (*lvts)[its].dUts0; }
  os << "\ndH*298,kJmol-1";
  for (its = 0; its < nts; its++) { os << "," << (*lvts)[its].dHts298; }
  os << "\ndS*298,JK-1mol-1";
  for (its = 0; its < nts; its++) { os << "," << (*lvts)[its].dSts298; }
  os << "\n\n";

  os.unsetf(ios::fixed); os.precision(7);
  os << "T[K]";
  if (lvts->calEql) { os << ",Keq"; }
  os << ",k,1000/T,ln(k),coord\n";
  for (icr = 0; icr < ncr; icr++) {
    T = lvts->vtstRes[icr].T;     kT = lvts->vtstRes[icr].kTST;
    Ke = lvts->vtstRes[icr].Keq;  crd = lvts->vtstRes[icr].coord;
    os << T;
    if (lvts->calEql) { os << "," << Ke; }
    os << "," << kT << "," << (1000. / T) << "," << log(kT)
       << "," << crd << endl;
  }
  os << "\n,T\ncoord";
  for (icr = 0; icr < ncr; icr++) { os << "," << lvts->vtstRes[icr].T; }
  os << endl;
  for (its = 0; its < nts; its++) {
    os << (*lvts)[its].coord;
    for (icr = 0; icr < ncr; icr++)
      { os << "," << (*lvts)[its].calcRes[icr].kTST; }
    os << endl;
  }
  os.precision(prec);
}

/*------------------------------------------------------------------------
     classes for variational transition state theory calculation
------------------------------------------------------------------------*/

varTState::varTState() { clear(); }               // --- (constructor) ---

void varTState::clear() {                                 // --- clear ---
  energySet = false;
  flname.erase();
  calcRes.clear();
}

listVTS::listVTS() { clear(); }                   // --- (constructor) ---

void listVTS::clear() {                                   // --- clear ---
  vector<varTState>::clear(); vtstRes.clear(); calEql = false;
}

